#!/usr/bin/env bash

###############################################################################
# Copyright (c) 2022 NVIDIA Corporation. All rights reserved.
#
# This Software is licensed under one of the following licenses:
#
# 1) under the terms of the "Common Public License 1.0" a copy of which is
#    available from the Open Source Initiative, see
#    http://www.opensource.org/licenses/cpl.php.
#
# 2) under the terms of the "The BSD License" a copy of which is
#    available from the Open Source Initiative, see
#    http://www.opensource.org/licenses/bsd-license.php.
#
# 3) under the terms of the "GNU General Public License (GPL) Version 2" a
#    copy of which is available from the Open Source Initiative, see
#    http://www.opensource.org/licenses/gpl-license.php.
#
# Licensee has the right to choose one of the above licenses.
#
# Redistributions of source code must retain the above copyright
# notice and one of the license notices.
#
# Redistributions in binary form must reproduce both the above copyright
# notice, one of the license notices in the documentation
# and/or other materials provided with the distribution.
###############################################################################

set -Eeu
trap cleanup SIGINT SIGTERM ERR EXIT
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" &>/dev/null && pwd -P)

cleanup() {
  trap - SIGINT SIGTERM ERR EXIT
  unset MLNX_SF_FILE
  unset MLNX_OVS_FILE
  unset SFC_CONF_FILE
  unset USER_PCI_ADDR
  unset NUM_SF
  unset ECPF0
  unset ECPF1
  unset REBOOT_OPTION
}


## Global variables
MLNX_SF_FILE="/etc/mellanox/mlnx-sf.conf"
MLNX_OVS_FILE="/etc/mellanox/mlnx-ovs.conf"
SFC_CONF_FILE="/etc/mellanox/sfc.conf"
UDEV_SCRIPT_PATH="/usr/lib/udev"
SF_RENAME_SCRIPT="auxdev-sf-netdev-rename"
SF_REP_RENAME_SCRIPT="sf-rep-netdev-rename"
OVS_MAPPING_TEMPLATE_FILE="sfc.conf.tmpl"
IFUP_HOOK_50="50-ifup-hooks"
IFUP_HOOKS_PATH="/etc/networkd-dispatcher/routable.d"
SYSCTL_CONF="50-sfc-hbn.conf"
SYSCTL_PATH="/lib/sysctl.d"
REBOOT_OPTION="false"
USER_PCI_ADDR=""
ECPF0=2
ECPF1=2
MIN_SF=4
MAX_SF=236
CPU_AFFINITY_LIST="0-2"

# NVCONFIG variables corresponding to the default number of 16 SFs
PF_BAR2_ENABLE=${PF_BAR2_ENABLE:-0}
PER_PF_NUM_SF=${PER_PF_NUM_SF:-1}
PF_TOTAL_SF=${PF_TOTAL_SF:-40}
PF_SF_BAR_SIZE=${PF_SF_BAR_SIZE:-10}
NUM_PF_MSIX_VALID=${NUM_PF_MSIX_VALID:-0}
PF_NUM_PF_MSIX_VALID=${PF_NUM_PF_MSIX_VALID:-1}
PF_NUM_PF_MSIX=${PF_NUM_PF_MSIX:-228}

distro=`lsb_release -i -s``lsb_release -r -s | tr -d '.'`
distro=${distro,,}

if [ "$EUID" -ne 0 ]
  then echo "Please run as root"
  exit
fi

usage() {
  cat << EOF
Usage: $(basename "${BASH_SOURCE[0]}") [-p0 | --ecpf0 <num>]  [-p1 | --ecpf1 <num>] [-h] [-v] [-r]
Generate SF configs and install OVS SFC configuration for HBN.

Available options:

-p0, --ecpf0         [Optional] Number of SFs on ECPF0 (between $(($MIN_SF-2)) and $(($MAX_SF-2)))
-p1, --ecpf1         [Optional] Number of SFs on ECPF1 (between $(($MIN_SF-2)) and $(($MAX_SF-2)))
-r,  --reboot        [Optional] Reboot after install is completed
-h,  --help          [Optional] Print this help and exit
-v,  --verbose       [Optional] Print script debug info

EOF
  exit
}

msg() {
  echo >&2 -e "${1-}"
}

die() {
  local msg=$1
  local code=${2-1} # default exit status 1
  msg "$msg"
  exit "$code"
}

is_number(){
  local string=$1
  case $string in
    ''|*[!0-9]*) die "$string : Not a valid number" ;;
    *) return ;;
  esac
}

validate_sf_number(){
  if ((${NUM_SF} < $MIN_SF || ${NUM_SF} > $MAX_SF)); then
    die "Invalid number of SFs: ${NUM_SF}. Total number of SFs should be between $MIN_SF and $MAX_SF"
  fi
}

validate_address(){
  pci_address_regex="^[a-fA-F0-9]{4}:[a-fA-F0-9]{2}\:[a-fA-F0-9]{2}\.[a-fA-F0-9]{1}$"
  [[ "$USER_PCI_ADDR" =~  $pci_address_regex ]] && msg "matched" || die "Bad pci address format"
}

validate_ecpf0(){
  is_number $ECPF0
  if ((${ECPF0} < $(($MIN_SF-2)) || ${ECPF0} > $(($MAX_SF-2)))); then
    die "Invalid number of ECPF0 SFs: ${ECPF0}. Please enter number between $(($MIN_SF-2)) and $(($MAX_SF-2))"
  fi
}

validate_ecpf1(){
  is_number $ECPF1
  if ((${ECPF1} < $(($MIN_SF-2)) || ${ECPF1} > $(($MAX_SF-2)))); then
    die "Invalid number of ECPF1 SFs: ${ECPF1}. Please enter number between $(($MIN_SF-2)) and $(($MAX_SF-2))"
  fi
}

parse_params() {
  # Local variables, if any

  while :; do
    case "${1-}" in
    -h | --help) usage ;;
    -v | --verbose) set -x ;;
    -p0 | --ecpf0)
      ECPF0="${2-}"
      validate_ecpf0
      shift
      ;;
    -p1 | --ecpf1)
      ECPF1="${2-}"
      validate_ecpf1
      shift
      ;;
    -r | --reboot)  REBOOT_OPTION="true" ;;
    -?*) die "Unknown option: $1" ;;
    *) break ;;
    esac
    shift
  done

  return 0
}

parse_params "$@"

add_sfs(){
  local num_ecpf0=$(($ECPF0-1))
  local num_ecpf1=$(($ECPF1-1))
  local pci_addr_base="0000:03:00."
  local mac_p0=$(cat /sys/class/net/p0/address)
  local mac_p1=$(cat /sys/class/net/p1/address)
  local sfaddcmd=""


  msg "Clearing existing config in $MLNX_SF_FILE";
  > ${MLNX_SF_FILE}

  # Using IEEE allocated mac address prefix for Nvidia -> 00:04:4b
  # https://www.ipchecktool.com/tool/macfinder?FORM_SUBMIT=ouiSearch&REQUEST_TOKEN=&oui=nvidia
  # https://gitlab.com/wireshark/wireshark/-/raw/master/manuf
  base_mac_ecpf0=$(printf '00:04:4b:%02x:%02x\n' $[RANDOM%256] $[RANDOM%256])
  base_mac_ecpf1=$(printf '00:04:4b:%02x:%02x\n' $[RANDOM%256] $[RANDOM%256])
  
  # Generating configs for SFs on ECPF0
  for i in `seq 0 ${num_ecpf0}`
  do
    if (($i == 0 )); then
        # The first one we treat special since it would be mapped to p0, so we use p0's mac address instead.
        sfaddcmd="/sbin/mlnx-sf --action create --device ${pci_addr_base}0 --sfnum $i --hwaddr $mac_p0 -t --cpu-list $CPU_AFFINITY_LIST"
    else
        mac_idx=$(printf '%02x' $i)
        sfaddcmd="/sbin/mlnx-sf --action create --device ${pci_addr_base}0 --sfnum $i --hwaddr $base_mac_ecpf0:$mac_idx -t --cpu-list $CPU_AFFINITY_LIST"
    fi
      
    msg "Adding sf conf: ${sfaddcmd}"
    echo ${sfaddcmd}  >> ${MLNX_SF_FILE}
  done

  # Generating configs for SFs on ECPF1
  for i in `seq 0 ${num_ecpf1}`
  do
      if (($i == 0 )); then
        # The first one we treat special since it would be mapped to p1, so we use p1's mac address instead.
        sfaddcmd="/sbin/mlnx-sf --action create --device ${pci_addr_base}1 --sfnum $i --hwaddr $mac_p1 -t --cpu-list $CPU_AFFINITY_LIST"
    else
        mac_idx=$(printf '%02x' $i)
        sfaddcmd="/sbin/mlnx-sf --action create --device ${pci_addr_base}1 --sfnum $i --hwaddr $base_mac_ecpf1:$mac_idx -t --cpu-list $CPU_AFFINITY_LIST"
    fi

    msg "Adding sf conf: ${sfaddcmd}"
    echo ${sfaddcmd}  >> ${MLNX_SF_FILE}
  done

  msg "==> SF configs saved in $MLNX_SF_FILE"


}

remove_auto_ovsbridge_creation(){
    if [ ! -f $MLNX_OVS_FILE ]; then
      die "File: $MLNX_OVS_FILE is missing. Cannot proceed. Aborting!"
    fi
    sed -i 's/CREATE_OVS_BRIDGES=\"yes\"/CREATE_OVS_BRIDGES=\"no\"/g' ${MLNX_OVS_FILE}
}

confirm_reboot(){
    read -p "==> Reboot option specified. Please confirm (y/n)? " answer
case ${answer:0:1} in
    y|Y )
        msg " Reboot confirmed. Proceeding..."
        reboot
        exit
    ;;
    * )
        msg "Skipping automatic reboot. Please reboot as early as possible for settings to take effect."
    ;;
esac
}

generate_ovs_sf_mapping(){
    tmpfile=$(mktemp)
    cat $SCRIPT_DIR/$OVS_MAPPING_TEMPLATE_FILE > ${tmpfile}

    local vfidx=0
    local num_ecpf0=$(($ECPF0-3))  # To account for SFs allocated to pf0hpf and p0
    local num_ecpf1=$(($ECPF1-3))  # To account for SFs allocated to pf1hpf and p1

    if [[ ${NUM_SF} == 4 ]] ; then
      # We do not have any mapping to generate.
      # Add closing paranthesis (for array), copy the file to correct location and return.
        echo ")" >> $tmpfile
        cat $tmpfile
        cp $tmpfile $SFC_CONF_FILE
        return
    fi

    # Generating mappings for SFs on ECPF0
    for i in `seq 0 ${num_ecpf0}`
    do
        echo "\"pf0vf${i}-pf0vf${i}_sf_r\"" >> $tmpfile
    done

    # Generating mappings for SFs on ECPF1
    for i in `seq 0 ${num_ecpf1}`
    do
        echo "\"pf1vf${i}-pf1vf${i}_sf_r\"" >> $tmpfile
    done
    
    # Add closing paranthesis.
    echo ")" >> $tmpfile

    msg "==> Generated Mapping file"
    cat $tmpfile

    cp $tmpfile $SFC_CONF_FILE
    rm -rf $tmpfile

}

install_sfc_service(){
    msg "==> Installing sfc systemd service..."
    cp $SCRIPT_DIR/sfc.service /etc/systemd/system/sfc.service
    msg "==> Finished installing sfc systemd service."
    msg "==> Enabling sfc systemd service"
    systemctl daemon-reload
    systemctl enable sfc
    systemctl enable sfc-state-propagation.service
}

copy_files(){
    msg "==> Start copying udev SF rename scripts"
    # Backup original files first
    cp $UDEV_SCRIPT_PATH/$SF_RENAME_SCRIPT{,.original}
    cp $SCRIPT_DIR/$SF_RENAME_SCRIPT $UDEV_SCRIPT_PATH/
    cp $UDEV_SCRIPT_PATH/$SF_REP_RENAME_SCRIPT{,.original}
    cp $SCRIPT_DIR/$SF_REP_RENAME_SCRIPT $UDEV_SCRIPT_PATH/
    msg "==> Finished copying udev SF rename scripts"

    msg "==> Start copying sysctl conf"
    mkdir -p $SYSCTL_PATH
    cp $SCRIPT_DIR/$SYSCTL_CONF $SYSCTL_PATH/
    msg "==> Finished copying sysctl conf file"

    msg "==> Start copying ovs sfc files"
    # We do not need to copy sfc.conf since it is done by the generate_ovs_sf_mapping function.
    mkdir -p /opt/mellanox/sfc
    cp $SCRIPT_DIR/sfc.sh  /opt/mellanox/sfc/
    chmod +x /opt/mellanox/sfc/sfc.sh
    msg "==> Finished copying ovs sfc files"
}

install_sfc_hbn_software() {
	msg "==> Install SFC HBN packages"
	apt install -y hbn-runtime
	msg "==> Finished install SFC HBN packages"
}

configure_mgmt_vrf() {
    msg "==> Start copying ifup hooks"
    # IFUP_HOOK_50 used to create mgmt VRF
    # mgmt VRF is supported by NetworkManager in Ubuntu22.04
    mkdir -p $IFUP_HOOKS_PATH
    cp $SCRIPT_DIR/$IFUP_HOOK_50 $IFUP_HOOKS_PATH/
    chmod +x $IFUP_HOOKS_PATH/$IFUP_HOOK_50
    msg "==> Finished copying ifup hooks files"
    msg "==> Start mgmt VRF configuration"
    # Remove cloud-init configuration to prevent netplan files creation
    rm -f /var/lib/cloud/seed/nocloud-net/network-config
    if [ -e /etc/netplan/50-cloud-init.yaml ]; then
      rm -f /etc/netplan/50-cloud-init.yaml
      netplan apply > /dev/null 2>&1 || true
    fi
    # Change oob_net0 configuration using networkd
    cat >> /etc/systemd/network/25-oob-net0.network << EOF
[Match]
Name=oob_net0

[Network]
VRF=mgmt
DHCP=ipv4

[DHCP]
UseDomains=true
EOF
    # Set tmfifo_net0 configuration using networkd
    cat >> /etc/systemd/network/25-tmfifo-net0.network << EOF
[Match]
Name=tmfifo_net0

[Network]
VRF=mgmt
Address=192.168.100.2/30

[Route]
Gateway=192.168.100.1
Metric=1025

[DHCP]
UseDomains=true
EOF
    # Create management VRF
    cat >> /etc/systemd/network/25-mgmt.network << EOF
[Match]
Name=mgmt

[Network]
Address=127.0.0.1/8
Address=::1/128
EOF

  cat >> /etc/systemd/network/25-mgmt.netdev << EOF
[NetDev]
Name=mgmt
Kind=vrf

[VRF]
TableId=1001
EOF
    # Disable containerd and kubelet services
    systemctl disable kubelet
    systemctl disable containerd
    systemctl disable docker
    systemctl disable ssh
    # Disable NetworkManager services
    systemctl disable NetworkManager.service
    systemctl disable NetworkManager-wait-online.service
    # Enable containerd and kubelet services for management VRF
    # The actual service files created by /lib/systemd/system-generators/systemd-vrf-generator
    mkdir -p /etc/systemd/system/vrf@mgmt.target.wants
    ln -snf /etc/systemd/system/containerd@.service /etc/systemd/system/vrf@mgmt.target.wants/containerd@mgmt.service
    ln -snf /etc/systemd/system/kubelet@.service /etc/systemd/system/vrf@mgmt.target.wants/kubelet@mgmt.service
    ln -snf /etc/systemd/system/docker@.service /etc/systemd/system/vrf@mgmt.target.wants/docker@mgmt.service
    sed -i -e 's@ -H fd://@@' -e 's/Wants=containerd.service/Wants=containerd@mgmt.service/' /lib/systemd/system/docker.service
    # Generate services files for management VRF
    /lib/systemd/system-generators/systemd-vrf-generator /run/systemd/generator /run/systemd/generator.early /run/systemd/generator.late
    # Start containerd, docker and kubelet services onthe first boot after the deployment
    cat > $SCRIPT_DIR/services_helper << EOF
#!/bin/bash

systemctl daemon-reload
systemctl enable ssh@mgmt.service
systemctl start ssh@mgmt.service
systemctl start containerd@mgmt.service
systemctl start kubelet@mgmt.service
systemctl start docker@mgmt.service
EOF
    chmod +x $SCRIPT_DIR/services_helper
    sed -i -e "/runcmd:/a\ \ - [ $SCRIPT_DIR/services_helper ]" /var/lib/cloud/seed/nocloud-net/user-data
    # Prevent from import_doca_telemetry.sh enabling kubelet and containerd services
    perl -ni -e 'print unless /systemctl/' /opt/mellanox/doca/services/telemetry/import_doca_telemetry.sh
    # Use systemd-resolved to keep /etc/resolv.conf up-to-date
    rm -f /etc/resolv.conf
    ln -snf /run/systemd/resolve/resolv.conf /etc/resolv.conf
    # Create .bash_aliases to include VRF in the shell prompt
    cat >> /root/.bash_aliases << 'EOF'
NS=$(ip netns identify)
[ -n "$NS" ] && NS=":${NS}"

VRF=$(ip vrf identify)
[ -n "$VRF" ] && VRF=":${VRF}"

PS1='${debian_chroot:+($debian_chroot)}\u@\h${NS}${VRF}:\w\$ '
EOF
    msg "==> Finished mgmt VRF configuration"
}

disable_netplan() {
  msg "==> Disable netplan configuration by cloud-init"
  echo "network: {config: disabled}" > /etc/cloud/cloud.cfg.d/99-custom-networking.cfg
  /bin/rm -f /etc/netplan/*
}

grub_cmd_update() {
	sed -i -r -e 's/^(GRUB_CMDLINE_LINUX=.*)\"/\1 cgroup_no_v1=net_prio,net_cls"/' /etc/default/grub
}

cx_nvconfig(){
    msg "==> Start CX device configuration"
    RC=0

    for dev in `lspci -nD -d 15b3: | grep 'a2d[26c]' | cut -d ' ' -f 1`
    do
        msg "==> Running mlxconfig for $dev"
        output=`mlxconfig -y -d $dev set \
            PF_BAR2_ENABLE=$PF_BAR2_ENABLE \
            PER_PF_NUM_SF=$PER_PF_NUM_SF \
            PF_TOTAL_SF=$PF_TOTAL_SF \
            PF_SF_BAR_SIZE=$PF_SF_BAR_SIZE \
            NUM_PF_MSIX_VALID=$NUM_PF_MSIX_VALID \
            PF_NUM_PF_MSIX_VALID=$PF_NUM_PF_MSIX_VALID \
            PF_NUM_PF_MSIX=$PF_NUM_PF_MSIX; \
            echo Status: $?`
        rc=${output:0-1}
        RC=$((RC+rc))
        msg "==> MLXCONFIG:"
        msg "$output"
        if [ $rc -ne 0 ]; then
            msg "==> Failed to configure CX device: $dev"
        fi
    done

    for dev in `lspci -nD -d 15b3: | grep 'a2d[26c]' | cut -d ' ' -f 1`
    do
        msg "==> Running mlxfwreset for $dev"
        output=`mlxfwreset -d $dev -y -l 3 --sync 1 r 2>&1; echo Status: $?`
        rc=${output:0-1}
        RC=$((RC+rc))
        if [ $rc -ne 0 ]; then
            msg "==> MLXFWRESET:"
            msg "$output"
            msg "==> Failed to reset CX device"
        fi
    done
    msg "==> Finished CX device configuration. Status: $RC"
    if [ $RC -ne 0 ]; then
        msg "==> Power cycle is required"
    fi
    return $RC
}

###########################################
###             Main                    ###
###########################################

install_sfc_hbn_software

NUM_SF=$(($ECPF0+$ECPF1))
validate_sf_number

remove_auto_ovsbridge_creation
add_sfs
generate_ovs_sf_mapping

copy_files
install_sfc_service
if [[ "X$distro" =~ Xubuntu2[02]04 ]]; then
  # Configure management VRF for Ubuntu 20.04 and 22.04
  configure_mgmt_vrf
  disable_netplan
fi
grub_cmd_update
cx_nvconfig
if [ "${REBOOT_OPTION}" = "true" ]; then
  confirm_reboot
fi
msg ""
msg "#####################################################################################"
msg "                            NOTICE                                                   "
msg "OVS SFC install completed. Please reboot at the earliest for settings to take effect."
msg "#####################################################################################"
