#!/usr/bin/env bash

###############################################################################
# Copyright (c) 2022 NVIDIA Corporation. All rights reserved.
#
# This Software is licensed under one of the following licenses:
#
# 1) under the terms of the "Common Public License 1.0" a copy of which is
#    available from the Open Source Initiative, see
#    http://www.opensource.org/licenses/cpl.php.
#
# 2) under the terms of the "The BSD License" a copy of which is
#    available from the Open Source Initiative, see
#    http://www.opensource.org/licenses/bsd-license.php.
#
# 3) under the terms of the "GNU General Public License (GPL) Version 2" a
#    copy of which is available from the Open Source Initiative, see
#    http://www.opensource.org/licenses/gpl-license.php.
#
# Licensee has the right to choose one of the above licenses.
#
# Redistributions of source code must retain the above copyright
# notice and one of the license notices.
#
# Redistributions in binary form must reproduce both the above copyright
# notice, one of the license notices in the documentation
# and/or other materials provided with the distribution.
###############################################################################

# We do not set safe bash options intentionally since some ovs commands do return non-zero exit code.
# set -euo pipefail  

msg() {
  echo >&2 -e "${1-}"
}

###  OVS functions
OVS_TIMEOUT=30   # seconds to timeout ovs operation
OFCTL_TIMEOUT=30  # Openflow timeout
VSCTL="/usr/bin/ovs-vsctl --timeout=${OVS_TIMEOUT}"
OFCTL="/usr/bin/ovs-ofctl --timeout=${OFCTL_TIMEOUT}"
BR_NAME=br-sfc
OF_PORT_INDEX=1

check_br(){
    ${VSCTL} br-exists $1
    retval=$?
    if [ ${retval} -ne 0 ]; then
        msg "Bridge $1 not present"
    fi
}

create_bridge() {
    local bridgename=$1
    ${VSCTL} --may-exist add-br ${bridgename} 
}

delete_bridge() {
    local bridgename=$1
    ${VSCTL} --if-exists del-br ${bridgename}
}

del_flows(){
    local bridge=${1:-br-sfc}
    ${OFCTL} del-flows ${bridge}
}

set_ofport_index(){
  local index=$(${VSCTL} --data=bare list bridge ${BR_NAME} | grep ofport_index | cut -d "=" -f2)
  if ! [ -z "${index}" ]; then 
    msg "ofport_index on bridge ${BR_NAME} is :${index}"
    OF_PORT_INDEX=${index}
  fi
}

initialize(){
    delete_bridge ovsbr1
    delete_bridge ovsbr2

    msg "==> Ensuring OVS Hardware offload is enabled"
    ovs_hw_offload=$(${VSCTL} get Open_vSwitch . other_config:hw-offload 2> /dev/null | tr -d '\"')
    if [ "X${ovs_hw_offload}" != "Xtrue" ]; then
        ${VSCTL} set Open_vSwitch . other_config:hw-offload=true
        /etc/init.d/openvswitch-switch restart
    fi
    msg "==> Completed OVS Hardware offload configuration"
          
    create_bridge ${BR_NAME}
    check_br ${BR_NAME}
    del_flows ${BR_NAME}
    set_ofport_index

}

# Add_port() adds given port to the given bridge and sets openflow `ofport_request` 
# index based on global OF_PORT_INDEX
add_port(){
    local port=$1
    local bridge=${2:-br-sfc}
    local index=$(${VSCTL} list interface $port 2>/dev/null | grep ofport_request | cut -d ":" -f2 | xargs)
    if ! [ -z "${index}" ]; then 
        msg "Port is already added to bridge ${BR_NAME} and configured with ofport_request=${index}"
        return
    fi
    msg "Adding port: ${port} to bridge: ${bridge} with index ${OF_PORT_INDEX}"
    ${VSCTL} --may-exist add-port ${bridge} ${port}  -- set interface ${port} ofport_request=${OF_PORT_INDEX} 
    OF_PORT_INDEX=$((${OF_PORT_INDEX}+1))
}

# Add_patch_flows() adds openflow rules to set up bidirectional flows between the given ports and bridge.
# The flows uses `ofport_request` index numbers instead of port names.
# These are obtained from the interface table.

add_patch_flow(){
    local port1=$1
    local port2=$2
    local bridge=${3:-br-sfc}

    check_br ${bridge}

    msg "Setting flows for bridge: ${bridge} with ports: ${port1} and ${port2}"
    local port1_idx
    local port2_idx
    
    port1_idx=$(${VSCTL} list interface ${port1} 2>/dev/null | grep ofport_request | cut -d ":" -f2 | xargs)
    port2_idx=$(${VSCTL} list interface ${port2} 2>/dev/null | grep ofport_request | cut -d ":" -f2 | xargs)

    ${OFCTL} add-flow ${bridge} table=0,priority=10,in_port=${port1_idx},actions=output:${port2_idx} 
    ${OFCTL} add-flow ${bridge} table=0,priority=10,in_port=${port2_idx},actions=output:${port1_idx} 
    
}

# Given two ports and optionally a OVS bridge, add_patch_flow()
# connects the ports to the bridge and setups openflow rules
# to map traffic from one port to the other and vice versa.
# Returns: none
add_patch(){

    local port1=$1
    local port2=$2
    local bridge=${3:-br-sfc}

    add_port ${port1} ${bridge}
    add_port ${port2} ${bridge}

    add_patch_flow ${port1} ${port2} ${bridge}
}


#################################
###         Main              ###
#################################

source /etc/mellanox/sfc.conf  # Get mappings

initialize   # Initialize bridges, setup indexes, cleanup flows etc.

msg "Bridge name: ${BR_NAME}"
for MAP in "${MAPPINGS[@]}"; do
    port1=${MAP%%-*}
    port2=${MAP##*-}
    echo "Port mapping : ${port1} <---> ${port2}"  
    add_patch ${port1} ${port2} ${BR_NAME}
done

# Store index for future.
${VSCTL} set bridge ${BR_NAME} external_ids:ofport_index=${OF_PORT_INDEX}
msg "SFC Completed"
date -u >> /tmp/sfc-activated
